@@ -199,6 +199,11 @@ struct sna_fill_op {
 			       const struct sna_fill_op *op,
 			       const BoxRec *box,
 			       int count);
+	fastcall void (*points)(struct sna *sna,
+			       const struct sna_fill_op *op,
+			       int16_t dx, int16_t dy,
+			       const DDXPointRec *points,
+			       int count);
 	void (*done)(struct sna *sna, const struct sna_fill_op *op);
 };
 
@@ -231,7 +236,10 @@ struct sna_render {
 			  int16_t msk_x, int16_t msk_y,
 			  int16_t dst_x, int16_t dst_y,
 			  int16_t w, int16_t h,
+			  unsigned flags,
 			  struct sna_composite_op *tmp);
+#define COMPOSITE_PARTIAL 0x1
+#define COMPOSITE_FALLBACK 0x80000000
 
 	bool (*check_composite_spans)(struct sna *sna, uint8_t op,
 				      PicturePtr dst, PicturePtr src,
@@ -318,9 +326,6 @@ struct sna_render {
 	} glyph[2];
 	pixman_image_t *white_image;
 	PicturePtr white_picture;
-#if HAS_PIXMAN_GLYPHS
-	pixman_glyph_cache_t *glyph_cache;
-#endif
 
 	uint16_t vb_id;
 	uint16_t vertex_offset;
@@ -661,8 +666,8 @@ bool sna_blt_composite(struct sna *sna,
 		       int16_t src_x, int16_t src_y,
 		       int16_t dst_x, int16_t dst_y,
 		       int16_t width, int16_t height,
-		       struct sna_composite_op *tmp,
-		       bool fallback);
+		       unsigned flags,
+		       struct sna_composite_op *tmp);
 bool sna_blt_composite__convert(struct sna *sna,
 				int x, int y,
 				int width, int height,
