@@ -39,12 +39,6 @@
 
 #include <mipict.h>
 
-#if 0
-#define __DBG(x) ErrorF x
-#else
-#define __DBG(x)
-#endif
-
 /* TODO: Emit unantialiased and MSAA triangles. */
 
 #ifndef MAX
@@ -65,12 +59,11 @@ static void _assert_pixmap_contains_box(PixmapPtr pixmap, BoxPtr box, const char
 	    box->x2 > pixmap->drawable.width ||
 	    box->y2 > pixmap->drawable.height)
 	{
-		ErrorF("%s: damage box is beyond the pixmap: box=(%d, %d), (%d, %d), pixmap=(%d, %d)\n",
-		       __FUNCTION__,
-		       box->x1, box->y1, box->x2, box->y2,
-		       pixmap->drawable.width,
-		       pixmap->drawable.height);
-		assert(0);
+		FatalError("%s: damage box is beyond the pixmap: box=(%d, %d), (%d, %d), pixmap=(%d, %d)\n",
+			   function,
+			   box->x1, box->y1, box->x2, box->y2,
+			   pixmap->drawable.width,
+			   pixmap->drawable.height);
 	}
 }
 #define assert_pixmap_contains_box(p, b) _assert_pixmap_contains_box(p, b, __FUNCTION__)
@@ -197,6 +190,9 @@ composite_aligned_boxes(struct sna *sna,
 		goto done;
 	}
 
+	if (op == PictOpClear && sna->clear)
+		src = sna->clear;
+
 	if (force_fallback ||
 	    !sna->render.composite(sna, op, src, NULL, dst,
 				   src_x,  src_y,
@@ -204,7 +200,7 @@ composite_aligned_boxes(struct sna *sna,
 				   clip.extents.x1,  clip.extents.y1,
 				   clip.extents.x2 - clip.extents.x1,
 				   clip.extents.y2 - clip.extents.y1,
-				   memset(&tmp, 0, sizeof(tmp)))) {
+				   COMPOSITE_PARTIAL, memset(&tmp, 0, sizeof(tmp)))) {
 		unsigned int flags;
 		pixman_box16_t *b;
 		int i, count;
@@ -213,8 +209,11 @@ composite_aligned_boxes(struct sna *sna,
 		     __FUNCTION__));
 
 		flags = MOVE_READ | MOVE_WRITE;
-		if (n == 1 && op <= PictOpSrc)
-			flags = MOVE_WRITE | MOVE_INPLACE_HINT;
+		if (op <= PictOpSrc) {
+			flags |= MOVE_INPLACE_HINT;
+			if (n == 1)
+				flags &= ~MOVE_READ;
+		}
 
 		if (!sna_drawable_move_region_to_cpu(dst->pDrawable, &clip, flags))
 			goto done;
@@ -257,13 +256,16 @@ composite_aligned_boxes(struct sna *sna,
 				RegionIntersect(&region, &region, &clip);
 				b = REGION_RECTS(&region);
 				count = REGION_NUM_RECTS(&region);
-				for (i = 0; i < count; i++) {
-					fbComposite(op, src, NULL, dst,
-						    src_x + b[i].x1 - boxes[0].x1,
-						    src_y + b[i].y1 - boxes[0].y1,
-						    0, 0,
-						    b[i].x1, b[i].y1,
-						    b[i].x2 - b[i].x1, b[i].y2 - b[i].y1);
+				if (sigtrap_get() == 0) {
+					for (i = 0; i < count; i++) {
+						fbComposite(op, src, NULL, dst,
+							    src_x + b[i].x1 - boxes[0].x1,
+							    src_y + b[i].y1 - boxes[0].y1,
+							    0, 0,
+							    b[i].x1, b[i].y1,
+							    b[i].x2 - b[i].x1, b[i].y2 - b[i].y1);
+					}
+					sigtrap_put();
 				}
 				pixman_region_fini(&region);
 				pixman_region_fini(&region);
@@ -793,52 +795,55 @@ composite_unaligned_boxes_inplace__solid(struct sna *sna,
 			continue;
 		}
 
-		RegionTranslate(&clip, dx, dy);
-		count = REGION_NUM_RECTS(&clip);
-		extents = REGION_RECTS(&clip);
-		while (count--) {
-			int16_t y1 = dy + pixman_fixed_to_int(t->top);
-			uint16_t fy1 = pixman_fixed_frac(t->top);
-			int16_t y2 = dy + pixman_fixed_to_int(t->bottom);
-			uint16_t fy2 = pixman_fixed_frac(t->bottom);
+		if (sigtrap_get() == 0) {
+			RegionTranslate(&clip, dx, dy);
+			count = REGION_NUM_RECTS(&clip);
+			extents = REGION_RECTS(&clip);
+			while (count--) {
+				int16_t y1 = dy + pixman_fixed_to_int(t->top);
+				uint16_t fy1 = pixman_fixed_frac(t->top);
+				int16_t y2 = dy + pixman_fixed_to_int(t->bottom);
+				uint16_t fy2 = pixman_fixed_frac(t->bottom);
 
-			DBG(("%s: t=(%d, %d), (%d, %d), extents (%d, %d), (%d, %d)\n",
-			     __FUNCTION__,
-			     pixman_fixed_to_int(t->left.p1.x),
-			     pixman_fixed_to_int(t->top),
-			     pixman_fixed_to_int(t->right.p2.x),
-			     pixman_fixed_to_int(t->bottom),
-			     extents->x1, extents->y1,
-			     extents->x2, extents->y2));
+				DBG(("%s: t=(%d, %d), (%d, %d), extents (%d, %d), (%d, %d)\n",
+				     __FUNCTION__,
+				     pixman_fixed_to_int(t->left.p1.x),
+				     pixman_fixed_to_int(t->top),
+				     pixman_fixed_to_int(t->right.p2.x),
+				     pixman_fixed_to_int(t->bottom),
+				     extents->x1, extents->y1,
+				     extents->x2, extents->y2));
 
-			if (y1 < extents->y1)
-				y1 = extents->y1, fy1 = 0;
-			if (y2 >= extents->y2)
-				y2 = extents->y2, fy2 = 0;
+				if (y1 < extents->y1)
+					y1 = extents->y1, fy1 = 0;
+				if (y2 >= extents->y2)
+					y2 = extents->y2, fy2 = 0;
 
-			if (y1 < y2) {
-				if (fy1) {
+				if (y1 < y2) {
+					if (fy1) {
+						lerp32_unaligned_box_row(pixmap, color, extents,
+									 t, dx, y1, 1,
+									 SAMPLES_Y - grid_coverage(SAMPLES_Y, fy1));
+						y1++;
+					}
+
+					if (y2 > y1)
+						lerp32_unaligned_box_row(pixmap, color, extents,
+									 t, dx, y1, y2 - y1,
+									 SAMPLES_Y);
+
+					if (fy2)
+						lerp32_unaligned_box_row(pixmap, color,  extents,
+									 t, dx, y2, 1,
+									 grid_coverage(SAMPLES_Y, fy2));
+				} else if (y1 == y2 && fy2 > fy1) {
 					lerp32_unaligned_box_row(pixmap, color, extents,
 								 t, dx, y1, 1,
-								 SAMPLES_Y - grid_coverage(SAMPLES_Y, fy1));
-					y1++;
+								 grid_coverage(SAMPLES_Y, fy2) - grid_coverage(SAMPLES_Y, fy1));
 				}
-
-				if (y2 > y1)
-					lerp32_unaligned_box_row(pixmap, color, extents,
-								 t, dx, y1, y2 - y1,
-								 SAMPLES_Y);
-
-				if (fy2)
-					lerp32_unaligned_box_row(pixmap, color,  extents,
-								 t, dx, y2, 1,
-								 grid_coverage(SAMPLES_Y, fy2));
-			} else if (y1 == y2 && fy2 > fy1) {
-				lerp32_unaligned_box_row(pixmap, color, extents,
-							 t, dx, y1, 1,
-							 grid_coverage(SAMPLES_Y, fy2) - grid_coverage(SAMPLES_Y, fy1));
+				extents++;
 			}
-			extents++;
+			sigtrap_put();
 		}
 
 		RegionUninit(&clip);
@@ -881,37 +886,40 @@ pixman:
 		pi.color = color;
 		pi.op = op;
 
-		count = REGION_NUM_RECTS(&clip);
-		extents = REGION_RECTS(&clip);
-		while (count--) {
-			int16_t y1 = pixman_fixed_to_int(t->top);
-			uint16_t fy1 = pixman_fixed_frac(t->top);
-			int16_t y2 = pixman_fixed_to_int(t->bottom);
-			uint16_t fy2 = pixman_fixed_frac(t->bottom);
+		if (sigtrap_get() == 0) {
+			count = REGION_NUM_RECTS(&clip);
+			extents = REGION_RECTS(&clip);
+			while (count--) {
+				int16_t y1 = pixman_fixed_to_int(t->top);
+				uint16_t fy1 = pixman_fixed_frac(t->top);
+				int16_t y2 = pixman_fixed_to_int(t->bottom);
+				uint16_t fy2 = pixman_fixed_frac(t->bottom);
 
-			if (y1 < extents->y1)
-				y1 = extents->y1, fy1 = 0;
-			if (y2 >= extents->y2)
-				y2 = extents->y2, fy2 = 0;
-			if (y1 < y2) {
-				if (fy1) {
+				if (y1 < extents->y1)
+					y1 = extents->y1, fy1 = 0;
+				if (y2 >= extents->y2)
+					y2 = extents->y2, fy2 = 0;
+				if (y1 < y2) {
+					if (fy1) {
+						pixsolid_unaligned_box_row(&pi, extents, t, y1, 1,
+									   SAMPLES_Y - grid_coverage(SAMPLES_Y, fy1));
+						y1++;
+					}
+
+					if (y2 > y1)
+						pixsolid_unaligned_box_row(&pi, extents, t, y1, y2 - y1,
+									   SAMPLES_Y);
+
+					if (fy2)
+						pixsolid_unaligned_box_row(&pi, extents, t, y2, 1,
+									   grid_coverage(SAMPLES_Y, fy2));
+				} else if (y1 == y2 && fy2 > fy1) {
 					pixsolid_unaligned_box_row(&pi, extents, t, y1, 1,
-								   SAMPLES_Y - grid_coverage(SAMPLES_Y, fy1));
-					y1++;
+								   grid_coverage(SAMPLES_Y, fy2) - grid_coverage(SAMPLES_Y, fy1));
 				}
-
-				if (y2 > y1)
-					pixsolid_unaligned_box_row(&pi, extents, t, y1, y2 - y1,
-								   SAMPLES_Y);
-
-				if (fy2)
-					pixsolid_unaligned_box_row(&pi, extents, t, y2, 1,
-								   grid_coverage(SAMPLES_Y, fy2));
-			} else if (y1 == y2 && fy2 > fy1) {
-				pixsolid_unaligned_box_row(&pi, extents, t, y1, 1,
-							   grid_coverage(SAMPLES_Y, fy2) - grid_coverage(SAMPLES_Y, fy1));
+				extents++;
 			}
-			extents++;
+			sigtrap_put();
 		}
 
 		RegionUninit(&clip);
@@ -1052,7 +1060,7 @@ composite_unaligned_boxes_inplace(struct sna *sna,
 {
 	if (!force_fallback &&
 	    (is_gpu(sna, dst->pDrawable, PREFER_GPU_SPANS) ||
-	     picture_is_gpu(sna, src))) {
+	     picture_is_gpu(sna, src, PREFER_GPU_SPANS))) {
 		DBG(("%s: fallback -- not forcing\n", __FUNCTION__));
 		return false;
 	}
@@ -1119,37 +1127,40 @@ composite_unaligned_boxes_inplace(struct sna *sna,
 			pi.bits = pixman_image_get_data(pi.mask);
 			pi.op = op;
 
-			count = REGION_NUM_RECTS(&clip);
-			extents = REGION_RECTS(&clip);
-			while (count--) {
-				int16_t y1 = pixman_fixed_to_int(t->top);
-				uint16_t fy1 = pixman_fixed_frac(t->top);
-				int16_t y2 = pixman_fixed_to_int(t->bottom);
-				uint16_t fy2 = pixman_fixed_frac(t->bottom);
+			if (sigtrap_get() == 0) {
+				count = REGION_NUM_RECTS(&clip);
+				extents = REGION_RECTS(&clip);
+				while (count--) {
+					int16_t y1 = pixman_fixed_to_int(t->top);
+					uint16_t fy1 = pixman_fixed_frac(t->top);
+					int16_t y2 = pixman_fixed_to_int(t->bottom);
+					uint16_t fy2 = pixman_fixed_frac(t->bottom);
 
-				if (y1 < extents->y1)
-					y1 = extents->y1, fy1 = 0;
-				if (y2 > extents->y2)
-					y2 = extents->y2, fy2 = 0;
-				if (y1 < y2) {
-					if (fy1) {
+					if (y1 < extents->y1)
+						y1 = extents->y1, fy1 = 0;
+					if (y2 > extents->y2)
+						y2 = extents->y2, fy2 = 0;
+					if (y1 < y2) {
+						if (fy1) {
+							pixmask_unaligned_box_row(&pi, extents, t, y1, 1,
+										  SAMPLES_Y - grid_coverage(SAMPLES_Y, fy1));
+							y1++;
+						}
+
+						if (y2 > y1)
+							pixmask_unaligned_box_row(&pi, extents, t, y1, y2 - y1,
+										  SAMPLES_Y);
+
+						if (fy2)
+							pixmask_unaligned_box_row(&pi, extents, t, y2, 1,
+										  grid_coverage(SAMPLES_Y, fy2));
+					} else if (y1 == y2 && fy2 > fy1) {
 						pixmask_unaligned_box_row(&pi, extents, t, y1, 1,
-									  SAMPLES_Y - grid_coverage(SAMPLES_Y, fy1));
-						y1++;
+									  grid_coverage(SAMPLES_Y, fy2) - grid_coverage(SAMPLES_Y, fy1));
 					}
-
-					if (y2 > y1)
-						pixmask_unaligned_box_row(&pi, extents, t, y1, y2 - y1,
-									  SAMPLES_Y);
-
-					if (fy2)
-						pixmask_unaligned_box_row(&pi, extents, t, y2, 1,
-									  grid_coverage(SAMPLES_Y, fy2));
-				} else if (y1 == y2 && fy2 > fy1) {
-					pixmask_unaligned_box_row(&pi, extents, t, y1, 1,
-								  grid_coverage(SAMPLES_Y, fy2) - grid_coverage(SAMPLES_Y, fy1));
+					extents++;
 				}
-				extents++;
+				sigtrap_put();
 			}
 
 			pixman_image_unref(pi.image);
@@ -1173,19 +1184,23 @@ composite_unaligned_boxes_inplace(struct sna *sna,
 			dy = (clip.extents.y2 - clip.extents.y1 + num_threads - 1) / num_threads;
 			num_threads = (clip.extents.y2 - clip.extents.y1 + dy - 1) / dy;
 
-			for (i = 1; i < num_threads; i++) {
-				thread[i] = thread[0];
-				thread[i].y1 = y;
-				thread[i].y2 = y += dy;
-				sna_threads_run(rectilinear_inplace_thread, &thread[i]);
-			}
+			if (sigtrap_get() == 0) {
+				for (i = 1; i < num_threads; i++) {
+					thread[i] = thread[0];
+					thread[i].y1 = y;
+					thread[i].y2 = y += dy;
+					sna_threads_run(i, rectilinear_inplace_thread, &thread[i]);
+				}
 
-			assert(y < clip.extents.y2);
-			thread[0].y1 = y;
-			thread[0].y2 = clip.extents.y2;
-			rectilinear_inplace_thread(&thread[0]);
+				assert(y < clip.extents.y2);
+				thread[0].y1 = y;
+				thread[0].y2 = clip.extents.y2;
+				rectilinear_inplace_thread(&thread[0]);
 
-			sna_threads_wait();
+				sna_threads_wait();
+				sigtrap_put();
+			} else
+				sna_threads_kill();
 
 			pixman_image_unref(thread[0].dst);
 			pixman_image_unref(thread[0].src);
